module Arxiv.Wrapper
  where

import Control.Applicative ((<$>))
import Control.Monad.Trans (liftIO)
import Control.Monad.Trans.Resource (MonadResource)
import Data.Conduit ((.|))
import Data.Function ((&))
import Data.List (intercalate)
import Debug.Trace (trace)
import Network.Api.Arxiv (Expression(..),  Field(..), (/*/), (/+/),(/-/))
import Network.HTTP.Conduit (parseRequest)
import Network.HTTP.Simple as HT
import Network.HTTP.Types.Status
import Text.HTML.TagSoup
import qualified Data.ByteString as B hiding (unpack) 
import qualified Data.ByteString.Char8 as B  (unpack) 
import qualified Data.Conduit as C
import qualified Data.Conduit.List as CL
import qualified Network.Api.Arxiv as Ax

------------------------------------------------------------
type Limit = Int
type Soup  = Tag String

------------------------------------------------------------

-- apiTotal_simple :: [String] -> IO Int

------------------------------------------------------------
-- | Simple Query search for terms in All fields
apiSimple :: Maybe Limit -> [String] -> IO [Result]
apiSimple limit query = execQuery limit $ simpleQuery query

-- | Complex Query search for terms in specific fields to be defined by Ax.Query
apiComplex :: Maybe Limit -> Ax.Query -> IO [Result]
apiComplex limit query = execQuery limit query

------------------------------------------------------------
simpleQuery :: [String] -> Ax.Query
simpleQuery xs = Ax.Query (Just $ Exp $ All xs) [] 0 25

complexQuery :: Ax.Query
complexQuery = 
  let au = Exp $ Au ["Aaronson"]
      t1 = Exp $ Ti ["quantum"]
      t2 = Exp $ Ti ["complexity"]
      x  = au /*/ (t1 /+/ t2)
   in Ax.Query { Ax.qExp   = Just $ Exp $ All ["nuclear fusion", "plasma"]
               , Ax.qIds   = []
               , Ax.qStart = 0
               , Ax.qItems = 50
               }


execQuery :: Maybe Limit -> Ax.Query -> IO [Result]
execQuery Nothing  q = C.runConduitRes (searchAxv q .| CL.consume)
execQuery (Just l) q = C.runConduitRes (searchAxv q .| CL.take l)
-- execQuery q = C.runConduitRes (searchAxv' q .| outSnk)

----------------------------------------------------------------------
-- Execute query and start a source
----------------------------------------------------------------------
searchAxv :: MonadResource m => Ax.Query -> C.ConduitT () Result m ()
searchAxv q = 
  let s = Ax.mkQuery q
   in do rsp <- HT.httpBS =<< liftIO (parseRequest s)
         case getResponseStatus rsp of
           (Status 200 _) -> getSoup (getResponseBody rsp) >>= results q
           st             -> error $ "Error:" ++ show st

----------------------------------------------------------------------
-- Consume page by page
----------------------------------------------------------------------
getSoup :: MonadResource m =>
           B.ByteString -> C.ConduitT () Result m [Soup]
getSoup b = concat <$> (C.yield b .| toSoup .| CL.consume)


----------------------------------------------------------------------
-- Receive a ByteString and yield Soup
----------------------------------------------------------------------
toSoup :: MonadResource m => C.ConduitT B.ByteString [Soup] m ()
toSoup = C.awaitForever (C.yield . parseTags . B.unpack)

----------------------------------------------------------------------
-- Yield all entries and fetch next page
----------------------------------------------------------------------
results :: MonadResource m =>
           Ax.Query -> [Soup] -> C.ConduitT () Result m ()
results q sp =
  Ax.forEachEntryM sp (C.yield . mkResult)
          >> searchAxv (Ax.nextPage q)

----------------------------------------------------------------------
-- Get data and format
----------------------------------------------------------------------
data Result = Result { doi              :: String
                     , url              :: String
                     , primaryCategory  :: Maybe Ax.Category
                     , categories       :: [Ax.Category]
                     , journal          :: String
                     , authors          :: [Ax.Author]
                     , publication_date :: String
                     , year             :: String
                     , title            :: String
                     , abstract         :: String
                     , total            :: Int
                     } deriving (Show)

mkResult :: [Soup] -> Result
mkResult sp = let doi'             = Ax.getDoi sp
                  url'             = Ax.getPdf sp
                  primaryCategory' = Ax.getPrimaryCategory sp
                  categories'      = Ax.getCategories      sp
                  journal'         = Ax.getJournal         sp
                  authors'         = Ax.getAuthors         sp
                  publication_date'= Ax.getPublished       sp
                  year'            = Ax.getYear            sp
                  title'           = Ax.getTitle           sp & clean'
                  abstract'        = Ax.getSummary         sp & clean'
                  total'           = Ax.totalResults    sp
               in ( Result doi' url'
                           primaryCategory' categories'
                           journal' authors'
                           publication_date' year'
                           title' abstract'
                           total'
                  )
  where clean' x = let x' = clean ['\n', '\r', '\t'] x
                    in if null x' then "Not found" else x'
        clean _ [] = []
        clean d (c:cs) | c `elem` d =  " " <> clean d cs
                       | otherwise  = c:clean d cs


----------------------------------------------------------------------
-- Sink results To print
----------------------------------------------------------------------
outSnk :: MonadResource m => C.ConduitT Result C.Void m ()
outSnk = C.awaitForever (liftIO . putStrLn . show)

